"""
Code for rotations of points
"""

import numpy as np
import SimpleITK as sitk
from scipy.spatial.transform import Rotation

from . import utils as ut


def define_euler_rotation(rx, ry, rz, degrees=True, order="xyz"):
    """
    Wrapper of scipy.spatial.transform.Rotation.from_euler

    Parameters
    ----------
    rx : float
        Angle to rotate about X.
    ry : float
        Angle to rotate about Y.
    rz : float
        Angle to rotate about Z.
    degrees : bool, optional
        Are the rotations in degrees? The default is True.
    order : str, optional
        Order of axes to transform as string. Default is 'xyz',
        meaning transform will happen x-->y-->z.

    Returns
    -------
    scipy.spatial.transform.Rotation
        Scipy 3D rotation object.
    """
    return Rotation.from_euler(order, [rx, ry, rz], degrees=True)


def rotate_about_and_translate(points, rotation, pivot, translation):
    """
    Rotates points about a pivot point, then apply translation.

    Parameters
    ----------
    points : (Nx3) numpy array
        Points to rotate. Each point gets its own row.
    rotation : scipy.spatial.transform.Rotation
        Rotation object.
    pivot : numpy.ndarray
        Point to rotate around.
    translation : numpy.ndarray
        Additional translation to apply to points.

    Returns
    -------
    numpy.ndarray
        Rotated points.
    """
    return rotate_about(points, rotation, pivot) + translation


def rotate_about(points, rotation, pivot):
    """
    Rotates points about a pivot point.

    Parameters
    ----------
    points : (Nx3) numpy array
        Points to rotate. Each point gets its own row.
    rotation : scipy.spatial.transform.Rotation
        Rotation object.
    pivot : numpy.ndarray
        Point to rotate around.

    Returns
    -------
    (Nx3) numpy array
            Rotated points

    """
    return rotation.apply(points - pivot) + pivot


def rotation_matrix_to_sitk(
    rotation, center=np.array((0, 0, 0)), translation=np.array((0, 0, 0))
):
    """
    Convert numpy array rotation matrix to sitk affine.

    Parameters
    ----------
    rotation : numpy.ndarray
        Matrix representing rotation matrix in three dimensions.
    center : numpy.ndarray, optional
        Vector representing center of rotation, default is origin.
    translation : numpy.ndarray, optional
        Vector representing translation of transform (after rotation), default
        is zero.

    Returns
    -------
    sitk.AffineTransform
        SITK transform with parameters matching the input object.
    """
    S = sitk.AffineTransform(3)
    S.SetMatrix(tuple(rotation.flatten()))
    S.SetTranslation(translation.tolist())
    S.SetCenter(center.tolist())
    return S


def sitk_to_rotation_matrix(S):
    """
    Convert sitk affine transform to numpy array rotation matrix.

    Parameters
    ----------
    S : sitk.AffineTransform
        Affine transform object.

    Returns
    -------
    numpy.ndarray
        Matrix representing rotation matrix in three dimensions.
    numpy.ndarray
        Translation vector.
    numpy.ndarray
        Center of rotation.
    """
    R = np.array(S.GetMatrix()).reshape((3, 3))
    translation = np.array(S.GetTranslation())
    center = np.array(S.GetCenter())
    return R, translation, center


def scipy_rotation_to_sitk(
    rotation, center=np.array((0, 0, 0)), translation=np.array((0, 0, 0))
):
    """
    Convert Scipy 'Rotation' object to equivalent sitk.

    Parameters
    ----------
    rotation : scipy.spatial.transform.Rotation
        Rotation object.
    center : numpy.ndarray, optional
        Center of rotation, default is origin.
    translation : numpy.ndarray, optional
        Translation vector, default is zero.

    Returns
    -------
    sitk.AffineTransform
        SITK transform with parameters matching the input object.
    """
    S = rotation_matrix_to_sitk(rotation.as_matrix(), center, translation)
    return S


def rotation_matrix_from_vectors(a, b):
    """
    Find rotation matrix to align a with b.

    Parameters
    ----------
    a : numpy.ndarray
        Vector to be aligned with b.
    b : numpy.ndarray
        Vector.

    Returns
    -------
    numpy.ndarray
        Rotation matrix such that `rotation_matrix @ a` is parallel to `b`.
    """
    # Follows Rodrigues` rotation formula
    # https://math.stackexchange.com/a/476311

    nd = a.shape[0]
    if nd != b.shape[0]:
        raise ValueError("a must be same size as b")
    na = ut.norm_vec(a)
    nb = ut.norm_vec(b)
    c = np.dot(na, nb)
    if c == -1:
        return -np.eye(nd)
    v = np.cross(na, nb)
    ax = ut.skew_symmetric_cross_product_matrix(v)
    rotation_matrix = np.eye(nd) + ax + ax @ ax * (1 / (1 + c))
    return rotation_matrix


def _rotate_mat_by_single_euler(mat, axis, angle):
    "Helper function that rotates a matrix by a single Euler angle"
    rotation_matrix = Rotation.from_euler(axis, angle).as_matrix().squeeze()
    return mat @ rotation_matrix


def roll(input_mat, angle):
    """
    Apply a rotation around the x-axis (roll/bank angle) to the input matrix.

    Parameters
    ----------
    input_mat : numpy.ndarray
        The input matrix to be rotated.
    angle : float
        The angle of rotation around the x-axis in radians.

    Returns
    -------
    numpy.ndarray
        The rotated matrix.
    """
    return _rotate_mat_by_single_euler(input_mat, "x", angle)


def pitch(input_mat, angle):
    """
    Apply a rotation around the y-axis (pitch/elevation angle) to the input
    matrix.

    Parameters
    ----------
    input_mat : numpy.ndarray
        The input matrix to be rotated.
    angle : float
        The angle of rotation around the y-axis in radians.

    Returns
    -------
    numpy.ndarray
        The rotated matrix.
    """
    return _rotate_mat_by_single_euler(input_mat, "y", angle)


def yaw(input_mat, angle):
    """
    Apply a rotation around the z-axis (yaw/heading angle) to the input matrix.

    Parameters
    ----------
    input_mat : numpy.ndarray
        The input matrix to be rotated.
    angle : float
        The angle of rotation around the z-axis in radians.

    Returns
    -------
    numpy.ndarray
        The rotated matrix.
    """
    return _rotate_mat_by_single_euler(input_mat, "z", angle)


def extract_angles(mat):
    """
    Extract the Euler angles (roll, pitch, yaw) from a rotation matrix.

    Parameters
    ----------
    mat : numpy.ndarray
        The rotation matrix from which to extract the Euler angles.

    Returns
    -------
    tuple of float
        The extracted Euler angles (roll, pitch, yaw) in radians.
    """
    return tuple(Rotation.from_matrix(mat).as_euler("xyz"))


def combine_angles(x, y, z):
    """
    Combine Euler angles (roll, pitch, yaw) into a rotation matrix.

    Parameters
    ----------
    x : float
        The roll angle (rotation around the x-axis) in radians.
    y : float
        The pitch angle (rotation around the y-axis) in radians.
    z : float
        The yaw angle (rotation around the z-axis) in radians.

    Returns
    -------
    numpy.ndarray
        The resulting rotation matrix.
    """
    return Rotation.from_euler("xyz", [x, y, z]).as_matrix().squeeze()


def make_homogeneous_transform(R, translation, scaling=None):
    """
    Combines a rotation matrix and translation into a homogeneous transform.

    Parameters
    ----------
    R : numpy.array(N,N)
        Rotation matrix.
    translation : numpy.array(N,)
        Translation vector.
    scaling : numpy.ndarray, optional
        Scaling factors, by default None.

    Returns
    -------
    numpy.ndarray
        Homogeneous transformation matrix.
    """
    N, M = R.shape
    if N != M:
        raise ValueError("R must be square")
    if N != translation.shape[0]:
        raise ValueError("R and translation must have same size")
    if scaling is not None and scaling.shape[0] != N:
        raise ValueError("scaling must have same size as R")

    if scaling is None:
        R_adj = R
    else:
        R_adj = np.diag(scaling) @ R
    R_homog = np.eye(N + 1)
    R_homog[0:N, 0:N] = R_adj
    R_homog[0:N, N] = translation
    return R_homog


def affine_and_translation_from_homogeneous(R_homog):
    """
    Extract rotation and translation from a homogeneous transform.

    Parameters
    ----------
    R_homog : numpy.ndarray
        Homogeneous transformation matrix.

    Returns
    -------
    numpy.ndarray
        Rotation matrix.
    numpy.ndarray
        Translation vector.
    """
    N, M = R_homog.shape
    if N != M:
        raise ValueError("R_homog must be square")
    R = R_homog[0:N, 0:N]
    translation = R_homog[0:N, N]
    return R, translation


def prepare_data_for_homogeneous_transform(pts):
    """
    Prepare points for homogeneous transformation.

    Parameters
    ----------
    pts : np.array(N,M) or np.array(M)
        array np.array(N,M) or np. poin(M)
        aeturns
    -------
    numpy.ndarray
        (M+1)-D points with 1 in the last position.
    """
    nd = pts.ndim
    if nd == 1:
        M = pts.shape[0]
        pts_homog = np.ones(M + 1)
        pts_homog[0:M] = pts
    elif nd == 2:
        N, M = pts.shape
        pts_homog = np.ones((N, M + 1))
        pts_homog[:, 0:M] = pts
    else:
        raise ValueError("pts must be 1D or 2D")
    return pts_homog


def extract_data_for_homogeneous_transform(pts_homog):
    """
    Extract points formatted for homogeneous transformation.

    Parameters
    ----------
    pts_homog : np.array(N,M+1) or np.array(M+1)
        (M+1)-D points with 1 in the last position.

    Returns
    -------
    np.array(N,M) or np.array(M)
        array of N M-D points.
    """
    nd = pts_homog.ndim
    if nd == 1:
        M = pts_homog.shape[0] - 1
        pts = pts_homog[0:M]
    elif nd == 2:
        N, M = pts_homog.shape
        pts = pts_homog[:, 0 : (M - 1)]  # noqa: E203
    else:
        raise ValueError("pts_homog must be 1D or 2D")
    return pts


def _apply_homogeneous_transform_to_transposed_pts(pts, R_homog):
    pts_homog = prepare_data_for_homogeneous_transform(pts)
    transformed_pts_homog = pts_homog @ R_homog.T
    return extract_data_for_homogeneous_transform(transformed_pts_homog)


def apply_affine(pts, affine_R, translation):
    """
    Apply an affine transformation to a set of points.

    Parameters
    ----------
    pts : numpy.ndarray
        The input points to be transformed.
    affine_R : numpy.ndarray
        The affine rotation matrix.
    translation : numpy.ndarray
        The translation vector.

    Returns
    -------
    numpy.ndarray
        The transformed points.
    """
    R_homog = make_homogeneous_transform(affine_R, translation)
    return _apply_homogeneous_transform_to_transposed_pts(pts, R_homog)


def invert_affine(affine_R, translation):
    """
    Invert an affine transformation.

    Parameters
    ----------
    affine_R : numpy.ndarray
        The affine rotation matrix.
    translation : numpy.ndarray
        The translation vector.

    Returns
    -------
    numpy.ndarray
        The inverted rotation matrix.
    numpy.ndarray
        The inverted translation vector.
    """
    R_inv = np.linalg.inv(affine_R)
    t_inv = -R_inv @ translation
    return R_inv, t_inv


def apply_inverse_affine(pts, affine_R, translation):
    """
    Apply the inverse of an affine transformation to a set of points.

    Parameters
    ----------
    pts : numpy.ndarray
        The input points to be transformed.
    affine_R : numpy.ndarray
        The affine rotation matrix.
    translation : numpy.ndarray
        The translation vector.

    Returns
    -------
    numpy.ndarray
        The transformed points.
    """
    R_inv, t_inv = invert_affine(affine_R, translation)
    return apply_affine(pts, R_inv, t_inv)


def affine_from_rotate_scale(R, scale=None):
    """
    Create an affine transformation matrix from a rotation matrix and scaling
    factors.

    Parameters
    ----------
    R : numpy.ndarray
        The rotation matrix.
    scale : numpy.ndarray, optional
        The scaling factors, by default None.

    Returns
    -------
    numpy.ndarray
        The affine transformation matrix.
    """
    if scale is None:
        scale = np.ones(3)
    R_adj = np.diag(scale) @ R
    return R_adj


def apply_rotate_translate(pts, R, translation):
    """
    Apply rotation and translation to a set of points.

    Parameters
    ----------
    pts : numpy.ndarray
        The input points to be transformed.
    R : numpy.ndarray
        The 3x3 rotation matrix.
    translation : numpy.ndarray
        The 3-element translation vector.

    Returns
    -------
    numpy.ndarray
        The transformed points.
    """
    return apply_affine(pts, R, translation)


def invert_rotate_translate(R, translation):
    """
    Compute the inverse rotation and translation.

    Parameters
    ----------
    R : numpy.ndarray
        The 3x3 rotation matrix. Must satisfy R.T @ R = I.
    translation : numpy.ndarray
        The 3-element translation vector.

    Returns
    -------
    tuple
        A tuple containing:
        - R_inv (numpy.ndarray): The transpose of the rotation matrix.
        - t_inv (numpy.ndarray): The inverse translation vector.

    Notes
    -----
    R is assumed to be a rotation matrix, but this function does not check
    that it is orthogonal. The caller is responsible for ensuring that R is
    a valid rotation matrix.
    """
    R_inv = R.T
    t_inv = -translation @ R
    return R_inv, t_inv


def create_homogeneous_from_euler_and_translation(rx, ry, rz, tx, ty, tz):
    """
    Create a homogeneous transformation matrix from Euler angles and
    translation.

    Parameters
    ----------
    rx : float
        Rotation angle around the x-axis in radians.
    ry : float
        Rotation angle around the y-axis in radians.
    rz : float
        Rotation angle around the z-axis in radians.
    tx : float
        Translation along the x-axis.
    ty : float
        Translation along the y-axis.
    tz : float
        Translation along the z-axis.

    Returns
    -------
    numpy.ndarray
        Homogeneous transformation matrix.
    """
    R = combine_angles(rx, ry, rz)
    t = np.array([tx, ty, tz])
    return make_homogeneous_transform(R, t)


def ras_to_lps_transform(R, translation=None):
    """
    Transforms a rotation matrix and translation vector from RAS to LPS
    coordinate system, or vice-versa.

    Parameters
    ----------
    R : numpy.ndarray
        A 3x3 rotation matrix.
    translation : numpy.ndarray, optional
        A 3-element translation vector. If None, a zero vector is used. Default
        is None.

    Returns
    -------
    numpy.ndarray
        The transformed 3x3 rotation matrix in LPS coordinate system.
    numpy.ndarray
        The transformed 3-element translation vector in LPS coordinate system.

    Raises
    ------
    ValueError
        If R is not a 3x3 matrix.
    """
    if R.shape != (3, 3):
        raise ValueError("R must be a 3x3 matrix")
    if translation is None:
        translation = np.zeros(3)
    T = make_homogeneous_transform(R, translation)
    ras2lps = np.diag([-1, -1, 1, 1])
    T_out = ras2lps @ T @ ras2lps
    R_out = T_out[:3, :3]
    translation_out = T_out[:3, 3]
    return R_out, translation_out


def compose_transforms(R_1, translation_1, *args):
    """
    Compose a series of rotation matrices and translation vectors.

    The first transformation is applied first, followed by the second
    transform, and so on.

    Parameters
    ----------
    R_1 : numpy.ndarray
        The initial rotation matrix.
    translation_1 : numpy.ndarray
        The initial translation vector.
    *args : tuple
        Additional rotation matrices and translation vectors. Must be provided
        in pairs (R_2, translation_2, ...). The first transform defined by R_1
        and translation_1 is applied first, followed by the R_2 and
        translation_2, R_2,and so on.

    Returns
    -------
    numpy.ndarray
        The composed rotation matrix.
    numpy.ndarray
        The composed translation vector.

    Raises
    ------
    ValueError
        If the number of additional arguments is not even.
    """
    nargs = len(args)
    if nargs == 0:
        return R_1, translation_1
    elif nargs >= 2:
        if nargs % 2 != 0:
            raise ValueError("Invalid number of arguments")
        R_2 = args[0]
        translation_2 = args[1]
        R = R_2 @ R_1
        translation = translation_2 + R_2 @ translation_1
        return compose_transforms(R, translation, *args[2:])
    else:
        raise ValueError("Invalid number of arguments")


def itk_to_slicer_transform(itk_transform):
    """
    Converts an ITK transform to a Slicer transform.

    This function converts a given ITK transform, which uses the LPS coordinate
    system, to a Slicer transform, which uses the RAS coordinate system. The
    ITK transform is assumed to be a 4x4 homogeneous transformation matrix.

    Parameters
    ----------
    itk_transform : numpy.ndarray
        A 4x4 homogeneous transformation matrix representing the ITK transform.

    Returns
    -------
    numpy.ndarray
        A 3x3 numpy.ndarray representing the rotation matrix of the Slicer
        transform.
    numpy.ndarray
        A 1x3 numpy.ndarray representing the translation vector of the Slicer
        transform.
    """
    R, translation = ras_to_lps_transform(
        itk_transform[:3, :3], itk_transform[:3, 3]
    )
    T = make_homogeneous_transform(R, translation)
    transform_to_parent_RAS = np.linalg.inv(T)
    return transform_to_parent_RAS[:3, :3], transform_to_parent_RAS[:3, 3]



"""
Code to find the rotation matrix to align a headframe to a set of holes.
"""

import itertools as itr
import logging

import numpy as np
import SimpleITK as sitk
from scipy import optimize as opt
from scipy.spatial.transform import Rotation

from . import measurement as mrmsr
from . import optimization as mropt
from . import rotations as rot
from . import sitk_volume as sv
from . import utils as ut
from .file_io import slicer_files as sf

logger = logging.getLogger(__name__)

lps_axes = dict(
    ap=np.array([0, 1, 0]), dv=np.array([0, 0, 1]), ml=np.array([1, 0, 0])
)
def_orient_names = ("vertical", "horizontal")
def_ap_names = ("anterior", "posterior")
def_orient_comparison_axes = dict(
    horizontal=lps_axes["dv"], vertical=lps_axes["ap"]
)
def_orient_axes_dict = {
    orient: lps_axes[direction]
    for orient, direction in dict(horizontal="ap", vertical="dv").items()
}
def_design_centers = dict(
    horizontal=dict(
        anterior=np.array([-6.34, np.nan, 2.5]),
        posterior=np.array([-5.04, np.nan, 1]),
    ),
    vertical=dict(
        anterior=np.array([-5.09, 3.209, np.nan]),
        posterior=np.array([-6.84, 9.909, np.nan]),
    ),
)
def_orient_indices = dict(horizontal=[0, 2], vertical=[0, 1])
def_hole_order = dict(
    horizontal=["anterior", "posterior"],
    vertical=["posterior", "anterior"],
)


def get_segmentation_pca(seg_img, seg_vals):
    """Finds first pca axis of segmentation for segments in set seg_vals

    For each value in seg_vals, this will find the indices of seg_arr equal to
    that value and de-mean it. The first PC will then be found of the
    concatenated centered groups of indices.

    Parameters
    ---------
    seg_arr : SimpleITK.Image
        Array with annotation values in each index
    seg_vals : iterable
        Set of values that each element of seg_arr will be compared to.

    Returns
    -------
    pc_axis : first pc axis of the indices separately centered for each
    """
    # Centers each segmentation value separately
    centered = []
    for seg_val in seg_vals:
        p = sv.find_points_equal_to(seg_img, seg_val)
        m = np.mean(p, axis=0)
        centered.append(p - m)
    gp = np.concatenate(centered, axis=0)
    return ut.get_first_pca_axis(gp)


def slices_centers_of_mass(
    img, seg_img, axis_dim, seg_val, slice_seg_thresh=1
):
    """Finds the center of mass of image slices along array dimension

    Iterates through `img` along dimension `axis_dim`, finds how many elements
    of `seg_img` are equal to `seg_val`, and if that number is greater than
    or equal to `slice_seg_thresh` calculates the center of mass of the masked
    `img` on that slice. Centers of mass are based on the physical points
    corresponding to each index in `seg_img` found with
    `transform_sitk_indices_to_physical_points`.

    Parameters
    ----------
    img : SimpleITK.Image
        Grayscale image used to calculate center of mass
    seg_img : SimpleITK.Image
        segmentation image used to select elements of `img`. The spatial
        information of `seg_img` will be used to determine where each element
        is in space.
    axis_dim : integer
        Axis along which `img` is sliced, in SimpleITK axis order
    seg_val : number
        elements of `img` will only be included in the center of mass if the
        corresponding element of `seg_img` is equal to `seg_val`
    slice_seg_thresh : integer
        center of mass along slices of `img` will only be calculated if at
        least `slice_seg_thresh` or more elements of `seg_img` are equal to
        `seg_val`. Default = 1.

    Returns
    -------
    com : np.ndarray (N x 3)
        center of mass for each slice of `img` meeting the criteria described
        above.
    """
    seg_arr = sitk.GetArrayViewFromImage(seg_img)
    arr = sitk.GetArrayViewFromImage(img)
    ndxs = ut.find_indices_equal_to(seg_arr, seg_val)
    ndxs_sitk = ndxs[:, ::-1]
    slice_ndxs = np.unique(ndxs_sitk[:, axis_dim])
    nmask_in_slice = np.array(
        [np.count_nonzero(ndxs_sitk[:, axis_dim] == x) for x in slice_ndxs]
    )
    sel_slice_ndxs = slice_ndxs[nmask_in_slice >= slice_seg_thresh]
    ndx_points = sv.transform_sitk_indices_to_physical_points(
        seg_img, ndxs_sitk
    )
    com = np.zeros((sel_slice_ndxs.size, 3))
    for i, slice_ndx in enumerate(sel_slice_ndxs):
        mask = ndxs_sitk[:, axis_dim] == slice_ndx
        np_ndx = tuple(ndxs[mask, :].T)
        sel_v = arr[np_ndx]
        com[i, :] = np.sum(
            sel_v[:, np.newaxis] * ndx_points[mask, :], axis=0
        ) / np.sum(sel_v)
    return com


def find_hole(img, seg_img, seg_val, sel_ndxs):
    """Find the center of a hole based on its segmentation value

    sel_ndxs is in sitk axis order!
    Returns sitk axis order
    """
    if seg_img.GetSize() != img.GetSize():
        raise ValueError("Image and segmentation must have the same shape")
    seg_arr = sitk.GetArrayViewFromImage(seg_img)
    arr = sitk.GetArrayViewFromImage(img)
    ndxs = ut.find_indices_equal_to(seg_arr, seg_val)
    if np.size(ndxs) == 0:
        return None
    ndx_points = sv.transform_sitk_indices_to_physical_points(
        seg_img,
        ndxs[:, [2, 1, 0]],  # convert to sitk axis order
    )
    np_ndx = tuple(ndxs.T)
    sel_v = arr[np_ndx]
    sum_sel = np.sum(sel_v)
    if sum_sel > 0:
        found_center = np.nan * np.ones(3)
        found_center[sel_ndxs] = (
            np.sum(sel_v[:, np.newaxis] * ndx_points, axis=0) / sum_sel
        )[sel_ndxs]
    else:
        return None
    return found_center


def find_holes_by_orientation(
    img,
    seg_img,
    seg_vals_dict,
    orient_indices=def_orient_indices,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
):
    """
    Find holes in an image by different orientations and anterior-posterior
    names.

    Parameters
    ----------
    img : ndarray
        The original image in which holes need to be found.
    seg_img : ndarray
        The segmented image that identifies different regions.
    seg_vals_dict : dict
        A dictionary where keys are orientation names and values are
        dictionaries.  These inner dictionaries map anterior-posterior names to
        segmentation values.
    orient_indices : dict, optional
        A dictionary where keys are orientation names and values are the
        indices used to find the holes in the image.
    orient_names : list of str, optional
        A list of orientation names.
    ap_names : list of str, optional
        A list of anterior-posterior names.

    Returns
    -------
    found_centers : dict
        A dictionary where keys are orientation names and values are
        dictionaries.  These inner dictionaries map anterior-posterior names to
        the centers of found holes.  If a hole is not found for a given
        orientation and anterior-posterior name, that entry will be missing.

    Notes
    -----
    This function iterates over the given orientations and anterior-posterior
    names to find the holes in the image using the `find_hole` function. The
    centers of the found holes are returned in a nested dictionary structure.
    """
    found_centers = {orient: dict() for orient in orient_names}
    for orient in orient_names:
        for ap in ap_names:
            if ap in seg_vals_dict[orient]:
                maybe_hole = find_hole(
                    img,
                    seg_img,
                    seg_vals_dict[orient][ap],
                    orient_indices[orient],
                )
                if maybe_hole is not None:
                    found_centers[orient][ap] = maybe_hole
    return found_centers


def find_hole_angles(
    centers_dict,
    hole_order=def_hole_order,
    orient_comparison_axis=def_orient_comparison_axes,
    orient_axis_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
):
    """
    Calculate angles between holes for each orientation.

    Parameters
    ----------
    centers_dict : dict
        Dictionary of hole centers for each orientation and anterior-posterior
        name.
    hole_order : dict, optional
        Dictionary defining the order of holes for each orientation, by default
        `def_hole_order`.
    orient_comparison_axis : dict, optional
        Dictionary of comparison axes for each orientation, by default
        `def_orient_comparison_axes`.
    orient_axis_dict : dict, optional
        Dictionary of axes for each orientation, by default
        `def_orient_axes_dict`.
    orient_names : list of str, optional
        List of orientation names, by default `def_orient_names`.
    ap_names : list of str, optional
        List of anterior-posterior names, by default `def_ap_names`.

    Returns
    -------
    centers_ang : dict
        Dictionary of calculated angles for each orientation.
    """
    centers_ang = dict()
    for orient in orient_names:
        if set(centers_dict[orient].keys()) == set(ap_names):
            centers_diff = (
                centers_dict[orient][hole_order[orient][0]]
                - centers_dict[orient][hole_order[orient][1]]
            )
            cd_nnan = centers_diff.copy()
            cd_nnan[np.isnan(cd_nnan)] = 0
            centers_ang[orient] = ut.signed_angle_rh(
                orient_comparison_axis[orient],
                cd_nnan,
                orient_axis_dict[orient],
            )
    return centers_ang


def estimate_hole_axis_from_segmentation(seg_img, seg_vals, reference_axis):
    """
    Estimate the axis of a hole from segmentation values.

    Parameters
    ----------
    seg_img : SimpleITK.Image
        The segmentation image.
    seg_vals : list
        List of segmentation values used to identify the hole.
    reference_axis : ndarray
        Reference axis vector, direction of returned axis will be flipped if
        the dot product between reference_axis and the found axis is negative

    Returns
    -------
    axis : ndarray
        Estimated axis of the hole.
    """
    # Estimate axis rotations from segment locations
    axis = get_segmentation_pca(seg_img, seg_vals)
    if np.dot(axis, reference_axis) < 0:  # pragma: no cover
        axis *= -1
    return axis


def estimate_hole_axes_from_segmentation_by_orientation(
    seg_img,
    seg_val_dict,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
):
    """
    Estimate hole axes from segmentation by orientation.

    Parameters
    ----------
    seg_img : SimpleITK.Image
        Segmentation image.
    seg_val_dict : dict
        Nested dictionary of segmentation values for each orientation and AP.
    orient_axes_dict : dict
        Nested dictionary of LPS vectors for each orientation and AP.
    orient_names : list of str
        List of orientation names.
    ap_names : list of str
        List of anterior-posterior names.

    Returns
    -------
    initial_axes : dict
        Dictionary of initial axes as a 3 element vector for each orientation.
    """
    initial_axes = dict()
    for orient in orient_names:
        seg_vals = [
            seg_val_dict[orient][ap]
            for ap in ap_names
            if ap in seg_val_dict[orient]
        ]
        initial_axes[orient] = estimate_hole_axis_from_segmentation(
            seg_img, seg_vals, orient_axes_dict[orient]
        )
    return initial_axes


def calculate_centers_of_mass_for_image_and_segmentation(
    img,
    seg_img,
    initial_axes,
    seg_vals_dict,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    slice_seg_thresh=1,
):
    """
    Calculate centers of mass for image and segmentation.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image used for calculating center of mass.
    seg_img : SimpleITK.Image
        The segmentation image used for selecting elements of `img`.
    initial_axes : dict
        Dictionary of initial axes for each orientation.
    seg_vals_dict : dict
        Nested dictionary of segmentation values for each orientation and
        anterior-posterior names.
    orient_lps_vector_dict : dict, optional
        Dictionary of LPS vectors for each orientation.
    orient_names : list of str, optional
        List of orientation names.
    ap_names : list of str, optional
        List of anterior-posterior names.
    slice_seg_thresh : int, optional
        Minimum number of segmentation elements in a slice required for center
        of mass calculation, by default 1.

    Returns
    -------
    coms : dict
        Nested dictionary of Nx3 NDArray centers of mass for each orientation
        and anterior-posterior name.
    """
    coms = dict()
    for orient in orient_names:
        coms[orient] = dict()
        axis_dim = np.nonzero(orient_axes_dict[orient])[0][0]
        r = rot.rotation_matrix_from_vectors(
            initial_axes[orient], orient_axes_dict[orient]
        )
        s = rot.rotation_matrix_to_sitk(r)
        sinv = s.GetInverse()
        seg_img_rs = sv.resample3D(
            seg_img, sinv, interpolator=sitk.sitkNearestNeighbor
        )
        img_rs = sv.resample3D(
            img, sinv, interpolator=sitk.sitkNearestNeighbor
        )
        for ap in ap_names:
            if ap in seg_vals_dict[orient]:
                com = slices_centers_of_mass(
                    img_rs,
                    seg_img_rs,
                    axis_dim,
                    seg_vals_dict[orient][ap],
                    slice_seg_thresh=slice_seg_thresh,
                )
                coms[orient][ap] = (
                    r.T @ com.T
                ).T  # rotate com to original location
    return coms


def estimate_axis_rotations_from_centers_of_mass(
    coms,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
):
    """
    Estimate axis rotations from centers of mass.

    Parameters
    ----------
    coms : dict
        Nested dictionary of Nx3 NDArray centers of mass for each orientation
        and anterior-posterior name.
    orient_axes_dict : dict, optional
        Dictionary of LPS vectors for each orientation.
    orient_names : list of str, optional
        List of orientation names.
    ap_names : list of str, optional
        List of anterior-posterior names.

    Returns
    -------
    orient_rotation_matrices : dict
        Dictionary of 3x3 rotation matrices for each orientation.
    axes : dict
        Dictionary of 3 element vector axes for each orientation.
    """
    orient_rotation_matrices = {orient: dict() for orient in orient_names}
    axes = dict()
    for orient in orient_names:
        # center the coms for each segment separately
        # and find the axis for this orientation
        ccoms = []
        for ap in ap_names:
            if ap in coms[orient]:
                com = coms[orient][ap]
                m = np.mean(com, axis=0)
                ccoms.append(com - m[np.newaxis, :])
        joined_ccoms = np.concatenate(ccoms, axis=0)
        tmp_axis = ut.get_first_pca_axis(joined_ccoms)
        if np.dot(tmp_axis, orient_axes_dict[orient]) < 0:
            tmp_axis *= -1

        # remove off-axis mean for each segment separately
        coms_deproj_centered = []
        for ap in ap_names:
            if ap in coms[orient]:
                com = coms[orient][ap]
                com_proj = ut.vector_rejection(com, tmp_axis)
                proj_m = np.mean(com_proj, axis=0)
                coms_deproj_centered.append(com - proj_m[np.newaxis, :])
        joined_deproj_ccoms = np.concatenate(coms_deproj_centered, axis=0)
        axis = ut.get_first_pca_axis(joined_deproj_ccoms)
        if np.dot(axis, orient_axes_dict[orient]) < 0:
            axis *= -1
        R = rot.rotation_matrix_from_vectors(axis, orient_axes_dict[orient])
        axes[orient] = R.T @ orient_axes_dict[orient]
        orient_rotation_matrices[orient] = R
    return orient_rotation_matrices, axes


def find_rotation_to_match_hole_angles(
    img,
    seg_img,
    initial_orient_rotation_matrices,
    axes,
    seg_vals_dict,
    design_centers=def_design_centers,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    orient_indices=def_orient_indices,
    hole_order=def_hole_order,
    orient_comparison_axis=def_orient_comparison_axes,
    n_iter=10,
):
    """
    Find rotation matrix to match hole angles.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image.
    seg_img : SimpleITK.Image
        The segmentation image.
    initial_orient_rotation_matrices : dict
        Dictionary of initial orientation rotation matrices indexed by
        orientation.
    axes : dict
        Dictionary of axes for each orientation.
    seg_vals_dict : dict
        Dictionary of segmentation values for each orientation and
        anterior-posterior names.
    design_centers : dict
        Dictionary of design centers for each orientation and
        anterior-posterior names.
    orient_axes_dict : dict
        Dictionary of LPS vectors for each orientation.
    orient_names : list of str
        List of orientation names.
    ap_names : list of str
        List of anterior-posterior names.
    orient_indices : dict
        Dictionary of orientation indices.
    hole_order : dict
        Dictionary of hole order (list of AP names) for each orientation.
    orient_comparison_axis : dict
        Dictionary of comparison axes for each orientation.
    n_iter : int, optional
        Number of iterations, by default 10.

    Returns
    -------
    R : ndarray
        Rotation matrix to align img with design centers.
    translation : ndarray
        Offsets to align img with design centers.
    """
    nhole = 0
    for orient in orient_names:
        for ap in ap_names:
            if ap in seg_vals_dict[orient]:
                nhole += 1
    # Start measuring hole location and orientation using the estimated set of
    # axes
    bases = np.zeros((3, 3))
    bases[:, 1] = axes["horizontal"]  # P
    bases[:, 2] = ut.norm_vec(
        ut.vector_rejection(axes["vertical"], bases[:, 1])
    )  # S
    bases[:, 0] = np.cross(bases[:, 1], bases[:, 2])  # L

    s_rot = (
        initial_orient_rotation_matrices["horizontal"] @ bases[:, 2]
    )  # rotated Superior axis
    rad = ut.signed_angle_rh(
        s_rot,
        lps_axes["dv"],
        lps_axes["ap"],
    )
    Rot_y = Rotation.from_rotvec(rad * lps_axes["ap"])
    R_y = Rot_y.as_matrix()
    R = R_y @ initial_orient_rotation_matrices["horizontal"]

    S_init = rot.rotation_matrix_to_sitk(R)
    S_init_inv = S_init.GetInverse()

    seg_img_current = sv.resample3D(
        seg_img, S_init_inv, interpolator=sitk.sitkNearestNeighbor
    )
    img_current = sv.resample3D(
        img, S_init_inv, interpolator=sitk.sitkNearestNeighbor
    )

    found_centers = find_holes_by_orientation(
        img_current,
        seg_img_current,
        seg_vals_dict,
        orient_indices,
        orient_names,
        ap_names,
    )
    found_centers_ang, design_centers_ang = [
        find_hole_angles(
            centers,
            hole_order,
            orient_comparison_axis,
            orient_axes_dict,
            orient_names,
        )
        for centers in (found_centers, design_centers)
    ]
    # TODO: verify this works when not all segments are present
    hole_diffs = np.full((4, 3), np.nan)
    for i, (orient, ap) in enumerate(itr.product(orient_names, ap_names)):
        if ap in found_centers[orient]:
            hole_diffs[i, :] = (
                design_centers[orient][ap] - found_centers[orient][ap]
            )
    translation = np.nanmean(hole_diffs, axis=0)

    usable_orients = list(found_centers_ang.keys())
    found_centers_curr = found_centers
    found_centers_ang_curr = found_centers_ang
    for iter_no in range(n_iter):
        for orient in usable_orients:
            ang_err = (
                design_centers_ang[orient] - found_centers_ang_curr[orient]
            )
            Rot_update = Rotation.from_rotvec(
                ang_err * orient_axes_dict[orient]
            )
            R_update = Rot_update.as_matrix()
            R = R_update @ R
            S = rot.rotation_matrix_to_sitk(R)
            S_inv = S.GetInverse()
            seg_img_current = sv.resample3D(
                seg_img, S_inv, interpolator=sitk.sitkNearestNeighbor
            )
            img_current = sv.resample3D(
                img, S_inv, interpolator=sitk.sitkNearestNeighbor
            )
            found_centers_curr = find_holes_by_orientation(
                img_current,
                seg_img_current,
                seg_vals_dict,
                orient_indices,
                orient_names,
                ap_names,
            )
            found_centers_ang_curr = find_hole_angles(
                found_centers_curr,
                hole_order,
                orient_comparison_axis,
                orient_axes_dict,
                orient_names,
            )
            hole_diffs = np.full((4, 3), np.nan)
            for i, (orient, ap) in enumerate(
                itr.product(orient_names, ap_names)
            ):
                if ap in found_centers_curr[orient]:
                    hole_diffs[i, :] = (
                        design_centers[orient][ap]
                        - found_centers_curr[orient][ap]
                    )
            translation = np.nanmean(hole_diffs, axis=0)
    return R, translation


def estimate_coms_from_image_and_segmentation(
    img,
    seg_img,
    seg_vals_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    orient_axes_dict=def_orient_axes_dict,
):
    """
    Estimate centers of mass (COMs) from image and segmentation.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image.
    seg_img : SimpleITK.Image
        The segmentation image.
    seg_vals_dict : dict
        Nested dictionary of segmentation values.
    orient_names : tuple of str, optional
        Tuple of orientation names, by default ("vertical", "horizontal").
    ap_names : tuple of str, optional
        Tuple of anterior-posterior names, by default
        ("anterior", "posterior").
    lps_axes : dict, optional
        Dictionary of LPS axes, by default dict(ap=np.array([0, 1, 0]),
        dv=np.array([0, 0, 1]), ml=np.array([1, 0, 0])).
    hole_orient_axis : dict, optional
        Dictionary of hole orientation axes, by default dict(horizontal="ap",
        vertical="dv").

    Returns
    -------
    coms : dict
        Dictionary of Nx3 NDArrays of centers of mass for each orientation.
    """
    # Estimate axis rotations from segment locations
    initial_axes = estimate_hole_axes_from_segmentation_by_orientation(
        seg_img, seg_vals_dict, orient_axes_dict, orient_names, ap_names
    )

    # Find centers of mass (COM) for slices perpendicular to initial axes
    coms = calculate_centers_of_mass_for_image_and_segmentation(
        img,
        seg_img,
        initial_axes,
        seg_vals_dict,
        orient_axes_dict,
        orient_names,
        ap_names,
    )
    return coms


def estimate_rotation_and_coms_from_image_and_segmentation(
    img,
    seg_img,
    seg_vals_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    orient_comparison_axis=def_orient_comparison_axes,
    design_centers=def_design_centers,  # Bregma-relative mms (LPS)
    orient_indices=def_orient_indices,
    hole_order=def_hole_order,
    orient_axes_dict=def_orient_axes_dict,
    n_iter=10,
):
    """
    Estimate rotation and centers of mass (COMs) from image and segmentation.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image.
    seg_img : SimpleITK.Image
        The segmentation image.
    seg_vals_dict : dict
        Dictionary of segmentation values for each orientation and
        anterior-posterior names.
    orient_names : list of str, optional
        List of orientation names, by default `def_orient_names`.
    ap_names : list of str, optional
        List of anterior-posterior names, by default `def_ap_names`.
    orient_comparison_axis : dict, optional
        Dictionary of comparison axes for each orientation, by default
        `def_orient_comparison_axes`.
    design_centers : dict, optional
        Dictionary of design centers for each orientation and
        anterior-posterior names, by default `def_design_centers`.
    orient_indices : dict, optional
        Dictionary of orientation indices, by default `def_orient_indices`.
    hole_order : dict, optional
        Dictionary of hole order for each orientation, by default
        `def_hole_order`.
    orient_axes_dict : dict, optional
        Dictionary of axes for each orientation, by default
        `def_orient_axes_dict`.
    n_iter : int, optional
        Number of iterations, by default 10.

    Returns
    -------
    coms : dict
        Dictionary of centers of mass for each orientation and
        anterior-posterior name.
    R : ndarray
        Rotation matrix.
    translation : ndarray
        Offsets for each hole.
    """

    coms = estimate_coms_from_image_and_segmentation(
        img,
        seg_img,
        seg_vals_dict,
        orient_names=orient_names,
        ap_names=ap_names,
        orient_axes_dict=orient_axes_dict,
    )

    # Estimate axis rotations from centers of mass
    (
        orient_rotation_matrices,
        axes,
    ) = estimate_axis_rotations_from_centers_of_mass(
        coms,
        orient_axes_dict=orient_axes_dict,
        orient_names=orient_names,
        ap_names=ap_names,
    )

    R, translation = find_rotation_to_match_hole_angles(
        img,
        seg_img,
        orient_rotation_matrices,
        axes,
        seg_vals_dict,
        design_centers=design_centers,
        orient_axes_dict=orient_axes_dict,
        orient_names=orient_names,
        ap_names=ap_names,
        orient_indices=orient_indices,
        hole_order=hole_order,
        orient_comparison_axis=orient_comparison_axis,
        n_iter=n_iter,
    )

    return coms, R, translation


def make_segment_dict(
    segment_info,
    segment_format=None,
    ap_names=def_ap_names,
    orient_names=def_orient_names,
    ignore_list=[],
):
    """
    Create a dictionary of segment values based on the provided segment
    information.

    Parameters
    ----------
    segment_info : dict
        A dictionary containing segment information.
    segment_format : str, optional
        The format string used to generate the segment keys. Defaults to None.
    ap_names : list, optional
        A list of names for the anterior-posterior (AP) segments. Defaults to
        def_ap_names.
    orient_names : list, optional
        A list of names for the orientation segments. Defaults to
        def_orient_names.
    ignore_list : list, optional
        A list of segment names to ignore. Defaults to an empty list.

    Returns
    -------
    dict
        A dictionary of segment values, organized by orientation and AP
        segment.

    """
    if segment_format is None:
        segment_format = "{}_{}"
    seg_vals = dict()
    used_ignores = {x: False for x in ignore_list}
    for orient in orient_names:
        seg_vals[orient] = dict()
        for ap in ap_names:
            key_name = segment_format.format(ap, orient)
            if key_name in segment_info:
                if key_name in used_ignores:
                    used_ignores[key_name] = True
                    continue
                seg_vals[orient][ap] = segment_info[key_name]
    if not all(used_ignores.values()):
        unused_ignores = [k for k, v in used_ignores.items() if not v]
        logger.warning(
            "Not all ignore segments were used. "
            f"Unused ignores: {unused_ignores}"
        )
    logger.debug(f"Found segments: {seg_vals}")
    return seg_vals


def segment_dict_from_seg_odict(
    seg_odict,
    segment_format=None,
    ap_names=def_ap_names,
    orient_names=def_orient_names,
    ignore_list=[],
):
    """
    Create a dictionary of segments from a segmentation ordered dictionary.

    Parameters
    ----------
    seg_odict : OrderedDict
        The segmentation ordered dictionary.
    segment_format : str, optional
        The format of the segment names. Defaults to None.
    ap_names : list, optional
        The list of anatomical plane names. Defaults to def_ap_names.
    orient_names : list, optional
        The list of orientation names. Defaults to def_orient_names.

    Returns
    -------
    dict
        A dictionary of segments, where the keys are the segment names and the
        values are the segment information.

    Raises
    ------
    ValueError
        If no segments are found in the segmentation ordered dictionary.
    """
    segment_info = sf.find_seg_nrrd_header_segment_info(seg_odict)
    segment_dict = make_segment_dict(
        segment_info, segment_format, ap_names, orient_names, ignore_list
    )
    if all([len(d) == 0 for d in segment_dict.values()]):
        raise ValueError(
            "No segments found. Is the key format {key_format} correct?"
        )
    return segment_dict


def _compress_each(mask, *args):
    """
    Compresses each iterable in `*args` using `mask`.

    Parameters
    ----------
    mask : iterable
        A boolean mask indicating which elements to keep.
    *args : iterable
        Variable number of iterables to compress.

    Returns
    -------
    list
        A list of iterables, each compressed using `mask`.

    Examples
    --------
    >>> mask = [True, False, True]
    >>> arg1 = [1, 2, 3]
    >>> arg2 = ['a', 'b', 'c']
    >>> _compress_each(mask, arg1, arg2)
    [(1, 3), ('a', 'c')]
    """
    return list(zip(*itr.compress(zip(*args), mask)))


def find_hf_rotation_from_seg_and_lowerplane(
    img,
    seg_img,
    seg_odict,
    plane_pts,
    segment_format=None,
    ap_names=def_ap_names,
    orient_names=def_orient_names,
    niter_rot=10,
    niter_com=50000,
    niter_com_plane=10000,
    xtol=1e-12,
    ignore_list=[],
):
    """
    Calculate the headframe rotation and translation from segmentation data and
    a lower plane.

    Parameters
    ----------
    img : ndarray
        The input image.
    seg_img : ndarray
        The segmented image.
    seg_odict : dict
        Dictionary containing segmentation information.
    plane_pts : ndarray
        Points defining the lower plane.
    segment_format : optional
        Format of the segment dictionary. Default is None.
    ap_names : list, optional
        List of anterior-posterior names. Default is `def_ap_names`.
    orient_names : list, optional
        List of orientation names. Default is `def_orient_names`.
    niter_rot : int, optional
        Number of iterations for the initial rotation estimation. Default is
        10.
    niter_com : int, optional
        Number of iterations for the optimization considering only holes.
        Default is 50000.
    niter_com_plane : int, optional
        Number of iterations for the optimization including the lower plane.
        Default is 10000.
    xtol : float, optional
        Tolerance for termination by change in the optimization. Default is
        1e-12.
    ignore_list : list, optional
        List of segment names to ignore. Default is an empty list.

    Returns
    -------
    theta0 : ndarray
        Initial guess for the rotation and translation parameters.
    output_holes_only : ndarray
        Optimized parameters considering only the headframe holes.
    output_all : ndarray
        Final optimized parameters including the lower plane.

    Notes
    -----
    The function performs the following steps:
    1. Converts the segmentation dictionary.
    2. Estimates centers of mass and initial rotation and translation.
    3. Retrieves design centers and hole locations.
    4. Prepares data for optimization.
    5. Optimizes the transformation considering only the headframe holes.
    6. Refines the optimization by including the lower plane.
    """
    segment_dict = segment_dict_from_seg_odict(
        seg_odict, segment_format, ap_names, orient_names, ignore_list
    )

    # Get centers of mass and initial guess at rotation and translation
    coms, R, translation = (
        estimate_rotation_and_coms_from_image_and_segmentation(
            img, seg_img, segment_dict, n_iter=niter_rot
        )
    )
    euler0 = Rotation.from_matrix(R).as_euler("xyz")
    theta0 = np.concatenate((euler0, translation))

    # Get design centers and hole locations
    pts1, pts2, names = mropt.get_headframe_hole_lines(
        insert_underscores=True, return_plane=True
    )

    # slice and dice data for optimization
    moving = []
    pts1l = []
    pts2l = []
    weights = []
    hole_mask = []
    group_err_funs = []
    for i, name in enumerate(names):
        if name == "plane":
            moving.append(plane_pts)
            pts1l.append(pts1[i, :])
            pts2l.append(pts2[i, :])
            npt_plane = plane_pts.shape[0]
            weights.append(np.full(npt_plane, 1 / npt_plane))
            hole_mask.append(False)
            group_err_funs.append(mrmsr.dist_point_to_plane)
        else:
            ap, orient = name.split("_")
            if orient in coms:
                ap_coms = coms[orient]
                if ap in ap_coms:
                    these_coms = ap_coms[ap]
                    npt = these_coms.shape[0]
                    moving.append(these_coms)
                    pts1l.append(pts1[i, :])
                    pts2l.append(pts2[i, :])
                    weights.append(np.full(npt, 1 / npt))
                    hole_mask.append(True)
                    group_err_funs.append(mrmsr.dist_point_to_line)
    hole_only_list_of_lists = _compress_each(
        hole_mask, pts1l, pts2l, moving, weights
    )

    # Optimize transform only considering holes in the headframe
    output_holes_only = opt.fmin(
        mropt.revised_error_rotate_compare_weighted_lines,
        theta0,
        args=tuple(hole_only_list_of_lists),
        xtol=xtol,
        maxfun=niter_com,
        retall=1,
    )
    R_holes_only, transl_holes_only = mropt.unpack_theta(output_holes_only[0])

    # Now include the lower plane
    output_all = opt.fmin(
        mropt.revised_error_rotate_compare_weighted_lines,
        output_holes_only[0],
        args=(pts1l, pts2l, moving, weights, group_err_funs),
        xtol=xtol,
        maxfun=niter_com_plane,
        retall=1,
    )
    R_all, transl_all = mropt.unpack_theta(output_all[0])

    return R, translation, R_holes_only, transl_holes_only, R_all, transl_all



import json

import numpy as np


def read_neuroglancer_annotation_layers(
    filename,
    layer_names=None,
    return_description=True,
):
    """
    Reads annotation layers from a Neuroglancer JSON file and returns points in
    physical coordinates.

    This function reads the annotation layers from a Neuroglancer JSON file and
    returns the points in physical coordinates. The points are scaled by the
    voxel spacing found in the JSON file and are in units described by the
    `units` return value. Optionally, it returns the descriptions of the
    annotations if they exist.

    Notes
    -----
    The points in the Neuroglancer annotation layers are assumed to be in the
    order z, y, x, t. Only the spatial dimensions z, y, x are returned, in that
    order.

    Parameters
    ----------
    filename : str
        Path to the Neuroglancer JSON file.
    layer_names : str or list of str or None, optional
        Names of annotation layers to extract. If None, auto-detects all
        annotation layers. Default is None.
    return_description : bool, optional
        If True, returns annotation descriptions alongside points. Default is
        True.

    Returns
    -------
    annotations : dict
        Dictionary of annotation coordinates, scaled by the values in the
        dimension information, for each layer. The coordinates are in units
        described by
        the `units` return value.
    units : list of str
        Units of each dimension (e.g., ['m', 'm', 'm']).
    descriptions : dict or None
        Dictionary of annotation descriptions for each layer. Returned only if
        `return_description` is True, otherwise None. If `return_description`
        is True and there is no description for a point, its value will be
        None.
    """
    data = _load_json_file(filename)
    spacing, units = _extract_spacing(data["dimensions"])

    layers = data.get("layers", [])
    layer_names = _resolve_layer_names(
        layers, layer_names, layer_type="annotation"
    )
    annotations, descriptions = _process_annotation_layers(
        layers,
        layer_names,
        spacing=spacing,
        return_description=return_description,
    )

    return annotations, units, descriptions


def get_neuroglancer_annotation_points(
    filename,
    layer_names=None,
    return_description=True,
):
    """
    Reads annotation layers from a Neuroglancer JSON file and returns points in
    voxel indices.

    This function reads the annotation layers from a Neuroglancer JSON file and
    returns the points in voxel indices. Optionally, it returns the
    descriptions of the annotations if they exist.

    Notes
    -----
    The points in the Neuroglancer annotation layers are assumed to be in the
    order z, y, x, t. Only the indices for z, y, x are returned, in that order.

    Parameters
    ----------
    filename : str
        Path to the Neuroglancer JSON file.
    layer_names : str or list of str or None, optional
        Names of annotation layers to extract. If None, auto-detects all
        annotation layers. Default is None.
    return_description : bool, optional
        If True, returns annotation descriptions alongside points. Default is
        True.

    Returns
    -------
    annotations : dict
        Dictionary of annotation coordinates for each layer.
    descriptions : dict or None
        Dictionary of annotation descriptions for each layer. Returned only if
        `return_description` is True, otherwise None.
    """
    data = _load_json_file(filename)

    layers = data.get("layers", [])
    layer_names = _resolve_layer_names(
        layers, layer_names, layer_type="annotation"
    )
    annotations, descriptions = _process_annotation_layers(
        layers,
        layer_names,
        return_description=return_description,
    )

    return annotations, descriptions


def _load_json_file(filename):
    """
    Loads and parses a JSON file.

    Parameters
    ----------
    filename : str
        Path to the JSON file.

    Returns
    -------
    dict
        Parsed JSON data.
    """
    with open(filename, "r", encoding="utf-8") as f:
        return json.load(f)


def _extract_spacing(dimension_data):
    """
    Extracts voxel spacing from the Neuroglancer file.

    Parameters
    ----------
    dimension_data : dict
        Neuroglancer JSON dimension data.

    Returns
    -------
    spacing : numpy.ndarray
        Voxel spacing in each dimension.
    units : list of str
        Units of each dimension (e.g., ['m', 'm', 'm']).

    Raises
    ------
    ValueError
        If the required dimensions ('z', 'y', 'x') are not present in the file.
    """
    keep_order = ["z", "y", "x"]
    dimension_set = set(dimension_data.keys())
    missing = set(keep_order) - dimension_set
    if missing:
        raise ValueError(
            "Neuroglancer file must contain z, y, and x dimensions, "
            f"but missing: {missing}."
        )
    spacing = []
    units = []
    for dim in keep_order:
        space, unit = dimension_data[dim]
        spacing.append(space)
        units.append(unit)
    return np.array(spacing, dtype=float), units


def _resolve_layer_names(layers, layer_names, layer_type):
    """
    Resolves layer names based on user input or auto-detects layers of the
    given type.

    Parameters
    ----------
    layers : list of dict
        Neuroglancer JSON layers.
    layer_names : str or list of str or None
        User-specified layer names or None to auto-detect.
    layer_type : str
        Type of layer to extract ('annotation' or 'probe').

    Returns
    -------
    list of str
        List of resolved layer names.

    Raises
    ------
    ValueError
        If the input `layer_names` is invalid.
    """
    if isinstance(layer_names, str):
        return [layer_names]
    if layer_names is None:
        return [
            layer["name"] for layer in layers if layer["type"] == layer_type
        ]
    if isinstance(layer_names, list):
        return layer_names
    raise ValueError(
        "Invalid input for layer_names. Expected a string, "
        "list of strings, or None."
    )


def _process_annotation_layers(
    layers,
    layer_names,
    spacing=None,
    return_description=True,
):
    """
    Processes annotation layers to extract points and descriptions.

    Parameters
    ----------
    layers : list of dict
        Neuroglancer JSON layers.
    layer_names : list of str
        Names of annotation layers to extract.
    spacing : numpy.ndarray or None, optional
        Voxel spacing for scaling. If None, no scaling is done. Default is
        None.
    return_description : bool, optional
        Whether to extract descriptions alongside points. Default is True.

    Returns
    -------
    annotations : dict
        Annotation points for each layer.
    descriptions : dict or None
        Annotation descriptions for each layer, or None if not requested.
    """
    annotations = {}
    descriptions = {} if return_description else None
    for layer_name in layer_names:
        layer = _get_layer_by_name(layers, layer_name)
        points, layer_descriptions = _process_layer_and_descriptions(
            layer,
            spacing=spacing,
            return_description=return_description,
        )
        annotations[layer_name] = points
        if return_description:
            descriptions[layer_name] = layer_descriptions

    return annotations, descriptions


def _get_layer_by_name(layers, name):
    """
    Retrieves a layer by its name.

    Parameters
    ----------
    layers : list of dict
        Neuroglancer JSON layers.
    name : str
        Layer name to retrieve.

    Returns
    -------
    dict
        Layer data.

    Raises
    ------
    ValueError
        If the layer is not found.
    """
    for layer in layers:
        if layer["name"] == name:
            return layer
    raise ValueError(f'Layer "{name}" not found in the Neuroglancer file.')


def _process_layer_and_descriptions(
    layer,
    spacing=None,
    return_description=True,
):
    """
    Processes layer points and descriptions.

    Parameters
    ----------
    layer : dict
        Layer data.
    spacing : numpy.ndarray or None, optional
        Voxel spacing for scaling. If None, no scaling is done. Default is
        None.
    return_description : bool, optional
        Whether to extract descriptions. Default is True.

    Returns
    -------
    points : numpy.ndarray
        Scaled and reordered points.
    descriptions : numpy.ndarray or None
        Descriptions, or None if not requested.

    Raises
    ------
    ValueError
        If the annotation points do not have 4 dimensions (z, y, x, t).
    """
    points = []
    annotations = layer.get("annotations", [])
    for annotation in annotations:
        point_arr = np.array(annotation.get("point", []), dtype=float)
        if point_arr.shape[0] != 4:
            raise ValueError(
                "Annotation points expected to have 4 dimensions "
                f"(z, y, x, t), but {point_arr.shape[0]} found."
            )
        points.append(point_arr[:3])  # Keep only the first three dimensions
    points = np.stack(points) if points else np.empty((0, 3), dtype=float)
    if spacing is not None:
        points = points * spacing

    if return_description:
        descriptions = [
            annotation.get("description", None) for annotation in annotations
        ]
        return points, np.array(descriptions, dtype=object)
    return points, None


def get_image_sources(filename):
    """
    Reads image source URL(s) from a Neuroglancer JSON file.

    Parameters
    ----------
    filename : str
        Path to the Neuroglancer JSON file.

    Returns
    -------
    image_sources : dict
        Dictionary mapping image layer names to their source URLs.
    """
    data = _load_json_file(filename)
    image_sources = {}
    for layer in data.get("layers", []):
        if layer.get("type") == "image" and "name" in layer:
            image_sources[layer["name"]] = layer.get("source", None)
    return image_sources



"""Module to deal with coordinate systems"""

import numpy as np


def find_coordinate_perm_and_flips(src: str, dst: str):  # noqa: C901
    """Determine how to convert between coordinate systems

    This function takes a source `src` and destination `dst` string specifying
    two coordinate systems, and finds the permutation and sign flip such that a
    source array can be transformed from its own coordinate system to the
    destination coordinate system by first applying the permutation and then
    multiplying the resulting array by the direction. That is, the input can
    be transformed to the desired coordinate system with the following code:
    `dst_array = direction * src_array[:, perm]`, where `direction` and `perm`
    are the returned values of this function.

    Coordinate system are defined by strings specifying how each axis aligns to
    anatomical directions, with each character belonging to the set 'APLRIS',
    corresponding to Anterior, Posterior, Left, Right, Inferior, Superior.

    An example string would be 'RAS' corresponding to Right, Anterior, Superior
    for the first, second, and third axes respectively. The axis increases in
    the direction indicated (i.e. 'R' means values are more positive as you
    move to the patient's right).

    Parameters
    ---------
    src: str
        String specifying a coordinate system, as described above
    dst: str
        String specifying a coordinate system, as described above

    Returns
    -------
    perm : np.ndarray(dtype=int16) (N)
        Permutation array used to convert the `src` coordinate system to the
        `dst` coordinate system
    direction: np.ndarray(dtype=int16) (N)
        Direction array used to multiply the `src` coordinate system after
        permutation into the `dst` coordinate system
    """
    nel = len(src)
    if len(dst) != nel:
        raise ValueError("Inputs should be the same length")
    src_u = src.upper()
    dst_u = dst.upper()
    basic_pairs = dict(R="L", A="P", S="I")
    pairs = dict()
    for k, v in basic_pairs.items():
        pairs[k] = v
        pairs[v] = k
    src_order = dict()
    for i, s in enumerate(src_u):
        if s not in pairs:
            raise ValueError(
                "Source direction '{}' not in R/L, A/P, or I/S".format(s)
            )
        if s in src_order or pairs[s] in src_order:
            raise ValueError("Source axis '{}' not unique".format(s))
        src_order[s] = i
    perm = -1 * np.ones(nel, dtype="int16")
    direction = np.zeros(nel, dtype="int16")
    dst_set = set()
    for i, d in enumerate(dst_u):
        if d not in pairs:
            raise ValueError(
                "Destination direction '{}' not in R/L, A/P, or I/S".format(d)
            )
        if d in dst_set or pairs[d] in dst_set:
            raise ValueError("Destination axis '{}' not unique".format(d))
        if d in src_order:
            perm[i] = src_order[d]
            direction[i] = 1
        elif pairs[d] in src_order:
            perm[i] = src_order[pairs[d]]
            direction[i] = -1
        else:
            raise ValueError(
                (
                    "Destination direction '{}' has "
                    + "no match in source directions '{}'"
                ).format(d, src_u)
            )
        dst_set.add(d)
    return perm, direction


def convert_coordinate_system(
    arr: np.ndarray, src_coord: str, dst_coord: str
) -> np.ndarray:
    """Converts points in one anatomical coordinate system to another

    This will permute and multiply the NxM input array `arr` so that N
    M-dimensional points in the coordinate system specified by `src_coord` will
    be transformed into the destination coordinate system specified by
    `dst_coord`. The current implementation does not allow the dimensions to
    change.

    Coordinate systems are defined by strings specifying how each axis aligns
    to anatomical directions, with each character belonging to the set
    'APLRIS', corresponding to Anterior, Posterior, Left, Right, Inferior,
    Superior, respectively.

    An example string would be 'RAS' corresponding to Right, Anterior, Superior
    for the first, second, and third axes respectively. The axis increases in
    the direction indicated (i.e. 'R' means values are more positive as you
    move to the patient's right).

    Parameters
    ----------
    arr : np.ndarray (N x M)
        N points of M dimensions (at most three)
    src_coord: str
        String specifying a coordinate system, as described above
    dst_coord: str
        String specifying a coordinate system, as described above

    Returns
    -------
    out : np.ndarray (N x M)
        The N input points transformed into the destination coordinate system
    """
    perm, direction = find_coordinate_perm_and_flips(src_coord, dst_coord)
    if arr.ndim == 1:
        out = arr[perm]
    else:
        out = arr[:, perm]
    out *= direction
    return out



"""
Tools specific to computing arc angles
"""

import numpy as np
from scipy.spatial.transform import Rotation

from aind_mri_utils.rotations import ras_to_lps_transform


def calculate_arc_angles(vec, degrees=True, invert_AP=True):
    """
    Calculate the arc angles for a given vector.

    Parameters
    ----------
    vec : array_like
        A 3-element vector with ML, AP, and DV components. Directions should be
        in RAS.

    Returns
    -------
    tuple of float
        The calculated arc angles in degrees. The first element is the angle
        around the x-axis, and the second element is the angle around the
        y-axis.  Returns None if the input vector is a zero vector.
    """
    if np.linalg.norm(vec) == 0:
        return None
    if np.dot(vec, [0, 0, 1]) < 0:
        vec = -vec
    nv = vec / np.linalg.norm(vec)
    rx = -np.arcsin(nv[1])  # Our convention is backwards from right hand rule
    ry = np.arctan2(nv[0], nv[2])
    if degrees:
        rx = np.rad2deg(rx)
        ry = np.rad2deg(ry)
    if invert_AP:
        rx = -rx
    return rx, ry


def transform_matrix_from_angles(
    AP, ML, rotation=0, invert_AP=True, invert_rotation=True
):
    """
    Create a transform from arc angles.

    Parameters
    ----------
    AP : float
        The angle around the x-axis (anterior-posterior).
    ML : float
        The angle around the y-axis (medial-lateral).
    rotation : float, optional
        The rotation angle around the z-axis (default is 0).
    invert_AP : bool, optional
        If True, invert the AP angle to correct for the non-right-handed
        convention (default is True).
    invert_rotation : bool, optional
        If True, invert the rotation angle (default is True).

    Returns
    -------
    numpy.ndarray
        The transformation matrix.

    Notes
    -----
    Our convention for spin about the x-axis (AP) and the z-axis (DV) is not
    right-handed, so use `invert_AP=True` and `invert_rotation=True`,
    respectively, to correct for this.
    """
    if invert_AP:
        AP = -AP
    if invert_rotation:
        rotation = -rotation
    euler_angles = np.array([AP, ML, rotation])
    R = (
        Rotation.from_euler("XYZ", euler_angles, degrees=True)
        .as_matrix()
        .squeeze()
    )
    return ras_to_lps_transform(R)[0]



import unittest

import numpy as np

from aind_mri_utils.optimization import (
    get_headframe_hole_lines,
    optimize_transform_labeled_lines,
    optimize_transform_labeled_lines_with_plane,
)


class OptimizationTest(unittest.TestCase):
    def test_get_headframe_hole_lines(self) -> None:
        """
        Tests for get_headframe_hole_lines
        """
        ant_hrz_hole_pts = np.array([[-6.34, 0, 2.5], [-6.34, 6.5, 2.5]])
        pts1, _, names = get_headframe_hole_lines(
            version="0.1",
            insert_underscores=False,
            coordinate_system="LPS",
            return_plane=False,
        )

        self.assertTrue(np.allclose(pts1[0, :], ant_hrz_hole_pts[0, :]))
        self.assertFalse("_" in names[0])
        self.assertTrue(len(names) == 4)

        pts1, pts2, names = get_headframe_hole_lines(
            version="0.1",
            insert_underscores=True,
            coordinate_system="RAS",
            return_plane=True,
        )
        self.assertTrue(
            np.array_equal(
                pts2[0, :],
                np.array(
                    [
                        -ant_hrz_hole_pts[1, 0],
                        -ant_hrz_hole_pts[1, 1],
                        ant_hrz_hole_pts[1, 2],
                    ]
                ),
            )
        )
        self.assertTrue("_" in names[0])
        self.assertTrue(len(names) == 5)

        # Test that value error is raised if bad version in passed
        self.assertRaises(ValueError, get_headframe_hole_lines, version=12)

    def test_optimize_transform_labeled_lines(self) -> None:
        """
        Tests optimize_transform_labeled_lines
        """
        # Generate some test data by just moving the 'ground truth' point
        # around
        pts1, pts2, _ = get_headframe_hole_lines()

        move_pts = np.vstack([pts1, pts2])
        move_pts[:, 2] = move_pts[:, 2] + 1
        labels = np.array(
            [
                0,
                1,
                2,
                3,
                0,
                1,
                2,
                3,
            ]
        )
        weights = np.array(
            [
                1,
                1,
                1,
                1,
                0.99,
                0.99,
                0.99,
                0.99,
            ]
        )  # To test that the weights are being used
        # Test with default weights input
        init = np.zeros((6,))
        trans, T_frame = optimize_transform_labeled_lines(
            init,
            pts1,
            pts2,
            move_pts,
            labels,
        )
        self.assertTrue(T_frame[-1] - 1 < 1e-6)
        self.assertTrue(trans[-1, -1] - 1 < 1e-6)

        # Test with non-default weights input
        init = np.zeros((6,))
        trans, T_frame = optimize_transform_labeled_lines(
            init,
            pts1,
            pts2,
            move_pts,
            labels,
            weights=weights,
            gamma=0.5,
            normalize=True,
        )
        self.assertTrue(T_frame[-1] - 1 < 1e-6)
        self.assertTrue(trans[-1, -1] - 1 < 1e-6)

    def test_optimize_transform_labeled_lines_with_plane(self) -> None:
        """
        Test optimize_transform_labeled_lines_with_plane
        """
        pts1, pts2, _ = get_headframe_hole_lines(
            version="0.1",
            insert_underscores=False,
            coordinate_system="LPS",
            return_plane=True,
        )

        move_pts = np.vstack([pts1[:4, :], pts2[:4, :], pts2[-1, :]])
        move_pts[:, 2] = move_pts[:, 2] + 1

        labels = np.array([0, 1, 2, 3, 0, 1, 2, 3, 4])
        weights = np.ones(
            len(labels),
        )

        pts_for_line = np.ones(pts1.shape[0], dtype=bool)
        pts_for_line[-1] = False
        # Test with all no weights added
        init = np.zeros((6,))
        trans, T_frame = optimize_transform_labeled_lines_with_plane(
            init, pts1, pts2, pts_for_line, move_pts, labels
        )

        self.assertTrue(T_frame[-1] - 1 < 1e-6)
        self.assertTrue(trans[-1, -1] - 1 < 1e-6)

        # Test with weights input
        init = np.zeros((6,))
        trans, T_frame = optimize_transform_labeled_lines_with_plane(
            init,
            pts1,
            pts2,
            pts_for_line,
            move_pts,
            labels,
            weights=weights,
            gamma=0.5,
            normalize=True,
        )

        self.assertTrue(T_frame[-1] - 1 < 1e-6)
        self.assertTrue(trans[-1, -1] - 1 < 1e-6)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import SimpleITK as sitk

from aind_mri_utils import headframe_rotation as hr


def add_cylinder(arr, center, radius, sel_ndx, ndx_range, value):
    """Add a cylinder to an array"""
    x, y, z = np.meshgrid(*map(np.arange, arr.shape), indexing="ij")
    ndxs = [x, y, z]
    all_axes = set([0, 1, 2])
    a, b = all_axes.difference([sel_ndx])
    mask = (ndxs[a] - center[0]) ** 2 + (ndxs[b] - center[1]) ** 2 < radius**2
    mask = mask & (
        (ndx_range[0] <= ndxs[sel_ndx]) & (ndxs[sel_ndx] < ndx_range[1])
    )
    arr[mask] = value
    return arr


def add_z_cylinder(arr, center, radius, ndx_range, value):
    """Add a cylinder to an array"""
    add_cylinder(arr, center, radius, 2, ndx_range, value)


def add_y_cylinder(arr, center, radius, ndx_range, value):
    """Add a cylinder to an array"""
    add_cylinder(arr, center, radius, 1, ndx_range, value)


def add_x_cylinder(arr, center, radius, ndx_range, value):
    """Add a cylinder to an array"""
    add_cylinder(arr, center, radius, 0, ndx_range, value)


def make_cylinders(img_size, cylinder_defs, vals):
    seg_arr = np.zeros(img_size[::-1], dtype="uint8")
    for cylinder, val in zip(cylinder_defs, vals):
        center, radius, sel_ndx, ndx_range = cylinder
        # Reminder: numpy indexing is (z, y, x) while SimpleITK uses (x, y, z)
        # Need to convert these sitk definitions to numpy
        center_np = np.array(center)[::-1]
        sel_ndx_np = 2 - sel_ndx
        add_cylinder(seg_arr, center_np, radius, sel_ndx_np, ndx_range, val)
    seg_img = sitk.GetImageFromArray(seg_arr)
    return seg_img


class HeadframeRotationTest(unittest.TestCase):
    sitk_test_img_size = (64, 64, 32)
    sitk_test_img_center = np.array([31.5, 31.5])
    # These are SITK indices!
    cylinder_defs = [
        # center, radius, sel_ndx, ndx_range
        # axis of cylinder is in `sel_ndx` direction
        # `ndx_range` determines the range of the cylinder
        #  in `sel_ndx` direction
        ((16, 16), 5, 2, (0, 16)),
        ((45, 45), 5, 2, (16, 32)),
        ((45, 10), 5, 1, (0, 16)),
        ((16, 20), 5, 1, (48, 64)),
    ]
    seg_vals = range(1, len(cylinder_defs) + 1)
    orient_names = ("vertical", "horizontal")
    ap_names = ("anterior", "posterior")
    seg_vals_dict = {
        "vertical": {a: v for a, v in zip(ap_names, range(1, 3))},
        "horizontal": {a: v for a, v in zip(ap_names, range(3, 5))},
    }

    def test_slices_center_of_mass(self) -> None:
        # Reminder: numpy indexing is (z, y, x) while SimpleITK uses (x, y, z)
        # (column major vs row major)
        # So a LPS image should have the THIRD axis be the L axis in numpy

        # convert between numpy and simpleITK indexing for size
        np_test_img_size = self.sitk_test_img_size[::-1]
        img_arr = np.zeros(np_test_img_size)
        seg_arr = np.zeros_like(img_arr, dtype="uint8")
        sigma = 0.5
        grids = []
        for i in range(1, 3):
            grids.append(np.linspace(-1, 1, np_test_img_size[i]))
        # backwards because of numpy indexing
        y, x = np.meshgrid(*grids, indexing="ij")
        dst = np.sqrt(x**2 + y**2)
        normal = 1 / (sigma * np.sqrt(2 * np.pi))
        exp_normal = 1 / (2 * sigma**2)
        img_arr[0, :, :] = normal * np.exp(exp_normal * -((dst) ** 2))
        seg_arr[0, :, :] = img_arr[0, :, :] > 0.5
        # Copy the first slice to the rest of the slices
        for i in range(1, np_test_img_size[0]):
            img_arr[i, :, :] = img_arr[0, :, :]
            seg_arr[i, :, :] = seg_arr[0, :, :]
        img = sitk.GetImageFromArray(img_arr)
        seg_img = sitk.GetImageFromArray(seg_arr)
        coms = hr.slices_centers_of_mass(img, seg_img, 2, 1, 5)
        self.assertEqual(coms.shape, (np_test_img_size[0], 3))
        for i in range(coms.shape[0]):
            self.assertTrue(
                np.allclose(coms[i, :2], self.sitk_test_img_center)
            )

    def test_get_segmentation_pca(self) -> None:
        seg_img = make_cylinders(
            self.sitk_test_img_size, self.cylinder_defs, self.seg_vals
        )
        axis = hr.get_segmentation_pca(
            seg_img, list(self.seg_vals_dict["vertical"].values())
        )
        self.assertTrue(np.allclose(axis, hr.lps_axes["dv"]))

    def test_hole_finding_and_orientation(self) -> None:
        seg_img = make_cylinders(
            self.sitk_test_img_size, self.cylinder_defs, self.seg_vals
        )
        img = make_cylinders(
            self.sitk_test_img_size,
            self.cylinder_defs,
            np.ones(len(self.seg_vals)),
        )

        this_val = self.seg_vals_dict["vertical"]["anterior"]
        these_orient_indices = hr.def_orient_indices["vertical"]
        hole = hr.find_hole(img, seg_img, this_val, these_orient_indices)
        self.assertTrue(np.isnan(hole[2]))
        self.assertTrue(np.allclose(hole[:2], list(self.cylinder_defs[0][0])))

        none_hole = hr.find_hole(
            img, seg_img, 5, these_orient_indices
        )  # 5 is not a seg value
        self.assertTrue(none_hole is None)

        bad_img = make_cylinders(
            self.sitk_test_img_size[::-1],
            self.cylinder_defs,
            np.ones(len(self.seg_vals)),
        )
        self.assertRaises(
            ValueError,
            hr.find_hole,
            bad_img,
            seg_img,
            this_val,
            these_orient_indices,
        )

        zero_img = make_cylinders(
            self.sitk_test_img_size,
            self.cylinder_defs,
            np.zeros(len(self.seg_vals)),
        )
        none_hole = hr.find_hole(
            zero_img, seg_img, this_val, these_orient_indices
        )
        self.assertTrue(none_hole is None)

        holes_dict = hr.find_holes_by_orientation(
            img,
            seg_img,
            self.seg_vals_dict,
        )
        self.assertTrue(
            np.allclose(
                holes_dict["vertical"]["anterior"][:2],
                list(self.cylinder_defs[0][0]),
            )
        )
        self.assertTrue(
            np.allclose(
                holes_dict["vertical"]["posterior"][:2],
                list(self.cylinder_defs[1][0]),
            )
        )
        self.assertTrue(
            np.allclose(
                holes_dict["horizontal"]["anterior"][[0, 2]],
                list(self.cylinder_defs[2][0]),
            )
        )
        self.assertTrue(
            np.allclose(
                holes_dict["horizontal"]["posterior"][[0, 2]],
                list(self.cylinder_defs[3][0]),
            )
        )

        centers_ang = hr.find_hole_angles(holes_dict)
        self.assertAlmostEqual(centers_ang["vertical"], -0.7853981633)
        self.assertAlmostEqual(centers_ang["horizontal"], 1.9028557943377)

        initial_axes = hr.estimate_hole_axes_from_segmentation_by_orientation(
            seg_img,
            self.seg_vals_dict,
        )
        self.assertTrue(
            np.allclose(initial_axes["vertical"], hr.lps_axes["dv"])
        )
        self.assertTrue(
            np.allclose(initial_axes["horizontal"], hr.lps_axes["ap"])
        )

        coms = hr.calculate_centers_of_mass_for_image_and_segmentation(
            img,
            seg_img,
            initial_axes,
            self.seg_vals_dict,
        )
        com_answer_dict = {
            "vertical": {
                "anterior": np.array([[16.0, 16.0, x] for x in range(16)]),
                "posterior": np.array(
                    [[45.0, 45.0, x] for x in range(16, 32)]
                ),
            },
            "horizontal": {
                "anterior": np.array([[45.0, x, 10.0] for x in range(16)]),
                "posterior": np.array(
                    [[16.0, x, 20.0] for x in range(48, 64)]
                ),
            },
        }
        for orient, com_answer_dict_orient in com_answer_dict.items():
            for ap, com_answer in com_answer_dict_orient.items():
                self.assertTrue(np.allclose(coms[orient][ap], com_answer))

        orient_rotation_matrices, axes = (
            hr.estimate_axis_rotations_from_centers_of_mass(coms)
        )
        self.assertTrue(
            np.allclose(orient_rotation_matrices["vertical"], np.eye(3))
        )
        self.assertTrue(
            np.allclose(orient_rotation_matrices["horizontal"], np.eye(3))
        )
        self.assertTrue(np.allclose(axes["vertical"], hr.lps_axes["dv"]))
        self.assertTrue(np.allclose(axes["horizontal"], hr.lps_axes["ap"]))

        test_centers = dict(
            vertical=dict(
                anterior=np.array([16, 16, np.nan]),
                posterior=np.array([45, 45, np.nan]),
            ),
            horizontal=dict(
                anterior=np.array([45, np.nan, 10]),
                posterior=np.array([16, np.nan, 20]),
            ),
        )
        R, offset = hr.find_rotation_to_match_hole_angles(
            img,
            seg_img,
            orient_rotation_matrices,
            axes,
            self.seg_vals_dict,
            design_centers=test_centers,
        )
        self.assertTrue(np.allclose(R, np.eye(3)))
        self.assertTrue(np.allclose(offset, np.zeros(3)))

        coms = hr.estimate_coms_from_image_and_segmentation(
            img, seg_img, self.seg_vals_dict
        )
        for orient, com_answer_dict_orient in com_answer_dict.items():
            for ap, com_answer in com_answer_dict_orient.items():
                self.assertTrue(np.allclose(coms[orient][ap], com_answer))

        coms, R, offset = (
            hr.estimate_rotation_and_coms_from_image_and_segmentation(
                img, seg_img, self.seg_vals_dict, design_centers=test_centers
            )
        )
        for orient, com_answer_dict_orient in com_answer_dict.items():
            for ap, com_answer in com_answer_dict_orient.items():
                self.assertTrue(np.allclose(coms[orient][ap], com_answer))
        self.assertTrue(np.allclose(R, np.eye(3)))
        self.assertTrue(np.allclose(offset, np.zeros(3)))


if __name__ == "__main__":
    unittest.main()



"""
Functions for saving and loading transforms using SimpleITK.
"""

import numpy as np
import SimpleITK as sitk

from .. import rotations as rot


def save_sitk_transform(
    filename,
    rotation_matrix,
    translation=None,
):
    """Save a rigid transform to a SimpleITK (sitk) transform file

    The current implementation assumes that rotations are applied as: y = Rx +
    t, where R is the rotation matrix, x is the input point, and t is the
    translation vector. These vectors are assumed to be column vectors.

    A SimpleITK transform is written to `filename` using `sitk.WriteTransform`.

    Parameters
    ----------
    filename : str
        The name of the file where the transform will be saved.
    rotation_matrix : np.ndarray
        The transform to save. See notes for supported shapes.
    translation : np.ndarray, optional
        An explicit translation vector to override any translation inferred
        from `rotation_matrix`. Defaults to None.

    Raises
    ------
    ValueError
        If the shape of `rotation_matrix` is not supported.

    Notes
    -----
    - If `rotation_matrix` is (6,), the function calculates the rotation matrix
    using `rot.combine_angles` for the first three values and assigns the last
    three values as the translation vector.
    - If `translation` is provided, it overrides any translation inferred from
    `rotation_matrix`.
    - Supported shapes for `rotation_matrix` include:
    - np.array of shape (6,): Interpreted as rotation angles (first 3 values)
    and translation (last 3 values). A rigid transform is constructed using
    `aind_mri_utils.optimization.create_rigid_transform`.
    - np.array of shape (4, 4): A homogeneous transformation matrix.
    - np.array of shape (3, 4): A rigid transform with a rotation
    matrix and translation vector.
    - np.array of shape (3, 3): A rotation matrix without translation.
    Translation defaults to a zero vector.

    """

    if len(rotation_matrix) == 6:
        R = rot.combine_angles(*rotation_matrix[:3])
        found_translation = rotation_matrix[3:]
    elif rotation_matrix.shape == (4, 4):
        found_translation = rotation_matrix[:3, 3]
        R = rotation_matrix[:3, :3]
    elif rotation_matrix.shape == (3, 4):
        R = rotation_matrix[:, :3]
        found_translation = rotation_matrix[:, 3]
    elif rotation_matrix.shape == (3, 3):
        R = rotation_matrix
        found_translation = np.zeros(3)
    else:
        raise ValueError("Invalid transform shape")
    if translation is not None:
        found_translation = translation
    A = rot.rotation_matrix_to_sitk(R, translation=found_translation)
    sitk.WriteTransform(A, filename)


def load_sitk_transform(filename, homogeneous=False, invert=False):
    """
    Convert a sitk transform file to a 4x3 numpy array.

    Parameters
    ----------
    filename : string
        filename to load from.
    homogeneous : bool, optional
        If True, return a 4x4 homogeneous transform matrix. Default is False.

    Returns
    -------
    R: np.array(N,M)
        Rotation matrix. For three dimensional transforms: np.array(3,3). If
        homogeneous: np.array(4, 4)
    translation: np.array(L,)
        Translation vector.
    center: np.array(L,)
        Center of rotation.
    """
    A = sitk.ReadTransform(filename)
    if invert:
        A = A.GetInverse()
    R, translation, center = rot.sitk_to_rotation_matrix(A)
    if homogeneous:
        if not np.allclose(center, 0):
            raise NotImplementedError(
                "homogeneous only valid for transforms with center at 0"
            )
        R = rot.make_homogeneous_transform(R, translation)
    return R, translation, center
